/*
 * IRC - Internet Relay Chat, ports/nt/os/service.c
 * Port library file for NT version of ircd.
 * (c) 1997 Thomas Helvey <tomh@inxpress.net>
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "../../../config/win32.h"
#ifdef RUN_AS_NT_SERVICE
#include <windows.h>
#include <winsvc.h>
#include <string.h>
#include <stdio.h>
#include "os.h"

static SERVICE_STATUS          IRCDStatus;
static SERVICE_STATUS_HANDLE   IRCDStatusHandle;  

VOID WINAPI IRCDStart(DWORD argc, LPTSTR* argv); 
VOID WINAPI IRCDCtrlHandler(DWORD opcode); 
DWORD IRCDInit(DWORD argc, LPTSTR* argv, DWORD *specificError);

extern int ircd_main_init(int argc, char** argv);
extern void main_loop();
extern void s_rehash();
extern void s_die();
extern void server_reboot();

static void SvcDebug(const char* str, DWORD status) 
{
  static char buf[1024]; 
  if (strlen(str) < 1000) {
    sprintf(buf, str, status); 
    OutputDebugString(buf);
  } 
}

VOID WINAPI IRCDCtrlHandler(DWORD opcode)
{
  DWORD status;  
  switch(opcode) {
  case IRCD_SERVICE_CONTROL_REHASH: 
    /*
     * Do whatever it takes to continue here.
     */
    s_rehash();
    break;
  case IRCD_SERVICE_CONTROL_RESTART: 
    /*
     * Do whatever it takes to restart here.
     */
    server_reboot();
    break;
  case SERVICE_CONTROL_STOP: 
    /* 
     * Do whatever it takes to stop here.
     */
    s_die();
    IRCDStatus.dwWin32ExitCode = 0; 
    IRCDStatus.dwCurrentState  = SERVICE_STOP_PENDING; 
    IRCDStatus.dwCheckPoint    = 0; 
    IRCDStatus.dwWaitHint      = 0;  
    if (!SetServiceStatus(IRCDStatusHandle, &IRCDStatus))
    { 
      status = GetLastError(); 
      SvcDebug("[ircd] SetServiceStatus error %ld\n",status); 
    }  
    SvcDebug("[ircd] Leaving ircd service\n",0); 
    return;
  case SERVICE_CONTROL_INTERROGATE: 
    /* 
     * Fall through to send current status.
     */
    break;  
  default: 
    SvcDebug("[ircd] Unrecognized opcode %ld\n", opcode);
  }
  /*
   * Send current status.
   */
  if (!SetServiceStatus(IRCDStatusHandle,  &IRCDStatus))
  { 
    status = GetLastError(); 
    SvcDebug("[ircd] SetServiceStatus error %ld\n", status);
  }
} 

VOID WINAPI IRCDStart(DWORD argc, LPTSTR* argv)
{
  DWORD status = 0;
  DWORD specificError = 0;

  IRCDStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS; 
  IRCDStatus.dwCurrentState = SERVICE_START_PENDING; 
  IRCDStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP | 
                                  SERVICE_ACCEPT_SHUTDOWN;
  IRCDStatus.dwWin32ExitCode = 0; 
  IRCDStatus.dwServiceSpecificExitCode = 0; 
  IRCDStatus.dwCheckPoint = 0; 
  IRCDStatus.dwWaitHint = 0;  
  
  IRCDStatusHandle = RegisterServiceCtrlHandler(TEXT("ircd"), IRCDCtrlHandler);  
  
  if (0 == IRCDStatusHandle)
  { 
    SvcDebug("[ircd] RegisterServiceCtrlHandler failed %d\n", GetLastError());
    return;
  }

  /*
   * Initialization code goes here.
   */
  status = IRCDInit(argc, argv, &specificError);

  /* 
   * Handle error condition
   */
  if (status != NO_ERROR)
  { 
    IRCDStatus.dwCurrentState  = SERVICE_STOPPED; 
    IRCDStatus.dwCheckPoint    = 0; 
    IRCDStatus.dwWaitHint      = 0; 
    IRCDStatus.dwWin32ExitCode = status; 
    IRCDStatus.dwServiceSpecificExitCode = specificError;

    SetServiceStatus(IRCDStatusHandle, &IRCDStatus); 
    return;
  }
  /*
   * Initialization complete - report running status. 
   */
  IRCDStatus.dwCurrentState = SERVICE_RUNNING; 
  IRCDStatus.dwCheckPoint   = 0; 
  IRCDStatus.dwWaitHint     = 0;  
  if (!SetServiceStatus(IRCDStatusHandle, &IRCDStatus))
  { 
    status = GetLastError(); 
    SvcDebug("[ircd] SetServiceStatus error %ld\n",status);
  }  
  /*
   * This is where the service does its work.
   */
  main_loop();

  SvcDebug("[ircd] Returning the Main Thread\n",0);
  return;
}
 
/* 
 * ircd initialization function.
 * This calls the ircd main function to setup the server and
 * pass command line parameters to it.
 */
DWORD IRCDInit(DWORD argc, LPTSTR* argv, DWORD* specificError)
{
  if ((*specificError = ircd_main_init(argc, argv)) != 0)
    return -1;
  return 0;
} 

int main(void)
{ 
  SERVICE_TABLE_ENTRY DispatchTable[] = { 
    { TEXT("ircd"), IRCDStart}, 
    { 0, 0 }     
  };  

  if (!StartServiceCtrlDispatcher(DispatchTable)) {
    SvcDebug("[ircd] StartServiceCtrlDispatcher error = %d\n", GetLastError());
    return -1;
  }
  return 0;
}  

#endif /* RUN_AS_NT_SERVICE */
